/*
 *  tiff.c
 */

// TIFF_BIGENDIAN_OS bigendian - TIFF standard, and native for
// Motorola, PowerPC, SPARC TIFF_LITENDIAN_OS little endian -
// non-standard TIFFs, easy to write on x86 systems, Alphas etc

#ifdef OS_GENERIC
#define TIFF_BIGENDIAN_OS
#endif OS_GENERIC

#ifdef OS_GENERICLINUX
#define TIFF_LITENDIAN_OS
#endif OS_GENERICLINUX

#ifdef OS_SOLARISAMD64
#define TIFF_LITENDIAN_OS
#endif OS_SOLARISAMD64

#ifdef OS_SOLARISSPARC
#define TIFF_BIGENDIAN_OS
#endif OS_SOLARISSPARC

#ifdef OS_MACOSXG4
#define TIFF_BIGENDIAN_OS
#endif OS_MACOSXG4

#ifdef OS_MACOSXG5
#define TIFF_BIGENDIAN_OS
#endif OS_MACOSXG5

#ifdef OS_MACOSXCORE
#define TIFF_LITENDIAN_OS
#endif OS_MACOSXCORE

// default TIFF_BIGENDIAN_OS if nothing matches

#ifndef TIFF_BIGENDIAN_OS
#ifndef TIFF_LITENDIAN_OS
#define TIFF_BIGENDIAN_OS
#endif
#endif

//
// TIFF Structures
//

typedef struct {
  UInt16 tCode;
  UInt16 dataType;
  unsigned int   nDataValues;
  unsigned int   pDataField;
} TagStructure;

// TIFF Colour table

typedef struct{
  UInt16 red[256];   // 0x0000 -> 0xFFFF, if input is 0-255 then byte double
  UInt16 green[256]; // 0x0000 -> 0xFFFF, ie 0xCA -> 0xCACA
  UInt16 blue[256];  // 0x0000 -> 0xFFFF
} TIFFCTab;

//======================================================================================
//
// Prototypes
//
//======================================================================================


int zls_ReadCtab(char * cfname, TIFFCTab * tCtab);

int zls_Write_tif8     (char * aFile, UInt8  * img, TIFFCTab * outTIFFCTab, int nxOut, int nyOut );   // 8 bit colour table, use gray ctab for grayscale

//======================================================================================
// Convenient wrapper to keep io errors and file io together, if needed
//======================================================================================


static inline int zlsio_write ( FILE * f, void * p, Size s );
static inline int zlsio_write ( FILE * f, void * p, Size s ){
  fwrite( p, s, 1, f);
  return( ferror ( f ) );
}


//======================================================================================
//
// TIFF Colour Table Read Routine
//
//======================================================================================


int zls_ReadCtab(char * cfname, TIFFCTab * tCtab){
  //
  // Read Text based  Color Tables
  // FORMAT:
  // line 1: fmt  (0 = decimal, 1= hex, 2 = float 0.0 -> 1.0)
  // line 2: componentSize  (8 or 16)
  // line 3: nComponents    (1 [grayscale] or 3 [rgb])
  // line 4: nEntries       (256)
  // line 5-260: index red green blue
  //
  
  int err = noErr;
  
  int colourFmt;
  int colourIndex;
  int componentSize;
  int nComponents;
  int nEntries;
  FILE * cfile;
  
  unsigned int entryIndex, red, green, blue, gray;
  float fred, fgreen, fblue, fgray;
  
  cfile = fopen(cfname, "r");
  
  if (cfile) {
    
    fscanf(cfile,"%d",&colourFmt);
    fscanf(cfile,"%d",&componentSize);
    fscanf(cfile,"%d",&nComponents);
    fscanf(cfile,"%d",&nEntries);
    
    if (colourFmt == 0){ // decimal entries
      
      if (componentSize == 8) {
        
        if (nComponents == 3) {
          for (colourIndex = 0; colourIndex<nEntries; colourIndex++){
            fscanf(cfile,"%d %d %d %d",&entryIndex, &red, &green, &blue);
            tCtab->red[entryIndex]   = (UInt16)(red   | ( red << 8));
            tCtab->green[entryIndex] = (UInt16)(green | ( green << 8));
            tCtab->blue[entryIndex]  = (UInt16)(blue  | ( blue << 8));
          }
        }
        
        if (nComponents == 1) {
          for (colourIndex = 0; colourIndex<nEntries; colourIndex++){
            fscanf(cfile,"%d %d",&entryIndex, &gray);
            tCtab->red[entryIndex]   = (UInt16)(gray   | (gray << 8));
            tCtab->green[entryIndex] = (UInt16)(gray   | (gray << 8));
            tCtab->blue[entryIndex]  = (UInt16)(gray   | (gray << 8));
          }
        }
        
      }
      
      if (componentSize == 16) {
        
        if (nComponents == 3) {
          for (colourIndex = 0; colourIndex<nEntries; colourIndex++){
            fscanf(cfile,"%d %d %d %d",&entryIndex, &red, &green, &blue);
            tCtab->red[entryIndex]   = (UInt16)(red);
            tCtab->green[entryIndex] = (UInt16)(green);
            tCtab->blue[entryIndex]  = (UInt16)(blue);
          }
        }
        
        if (nComponents == 1) {
          for (colourIndex = 0; colourIndex<nEntries; colourIndex++){
            fscanf(cfile,"%d %d",&entryIndex, &gray);
            tCtab->red[entryIndex]   = (UInt16)(gray);
            tCtab->green[entryIndex] = (UInt16)(gray);
            tCtab->blue[entryIndex]  = (UInt16)(gray);
          }
        }
        
      }
    }
    
    if (colourFmt == 1){ // hex entries
      
      if (componentSize == 8) {
        
        if (nComponents == 3) {
          for (colourIndex = 0; colourIndex<nEntries; colourIndex++){
            fscanf(cfile,"%d %x %x %x",&entryIndex, &red, &green, &blue);
            tCtab->red[entryIndex]   = (UInt16)(red   |(red << 8));
            tCtab->green[entryIndex] = (UInt16)(green |(green << 8));
            tCtab->blue[entryIndex]  = (UInt16)(blue  |(blue << 8));
          }
        }
        
        if (nComponents == 1) {
          for (colourIndex = 0; colourIndex<nEntries; colourIndex++){
            fscanf(cfile,"%d %x",&entryIndex, &gray);
            tCtab->red[entryIndex]   = (UInt16)(gray   | (gray << 8));
            tCtab->green[entryIndex] = (UInt16)(gray   | (gray << 8));
            tCtab->blue[entryIndex]  = (UInt16)(gray   | (gray << 8));
          }
        }
        
      }
      
      if (componentSize == 16) {
        
        if (nComponents == 3) {
          for (colourIndex = 0; colourIndex<nEntries; colourIndex++){
            fscanf(cfile,"%d %x %x %x",&entryIndex, &red, &green, &blue);
            tCtab->red[entryIndex]   = (UInt16)(red);
            tCtab->green[entryIndex] = (UInt16)(green);
            tCtab->blue[entryIndex]  = (UInt16)(blue);
          }
        }
        
        if (nComponents == 1) {
          for (colourIndex = 0; colourIndex<nEntries; colourIndex++){
            fscanf(cfile,"%d %x",&entryIndex, &gray);
            tCtab->red[entryIndex]   = (UInt16)(gray);
            tCtab->green[entryIndex] = (UInt16)(gray);
            tCtab->blue[entryIndex]  = (UInt16)(gray);
          }
        }
        
      }
    }
    
    if (colourFmt == 2){ // float entries
      
      // ignore component size and map 0 - 1 to 0-65535  
      
      if (nComponents == 3) {
        for (colourIndex = 0; colourIndex<nEntries; colourIndex++){
          fscanf(cfile,"%d %f %f %f",&entryIndex, &fred, &fgreen, &fblue);
          tCtab->red[entryIndex]   = (UInt16)(fred   * 0x0000ffff);
          tCtab->green[entryIndex] = (UInt16)(fgreen * 0x0000ffff);
          tCtab->blue[entryIndex]  = (UInt16)(fblue  * 0x0000ffff);
        }
      }
      
      if (nComponents == 1) {
        for (colourIndex = 0; colourIndex<nEntries; colourIndex++){
          fscanf(cfile,"%d %f",&entryIndex, &fgray);
          tCtab->red[entryIndex]   = (UInt16)(fgray  * 0x0000ffff);
          tCtab->green[entryIndex] = (UInt16)(fgray  * 0x0000ffff);
          tCtab->blue[entryIndex]  = (UInt16)(fgray  * 0x0000ffff);
        }
      }
      
    }
    
    err      = ferror (cfile); // one error check for whole write sequence
    fclose(cfile);
  } else {err = -1;}
  
  return(err);
  
}

//========================================
//
// TIFF File Routine
//
//========================================


int zls_Write_tif8     (char * aFile, UInt8 * img, TIFFCTab * outTIFFCTab, int nxOut, int nyOut ){

  int err = 0;
  //
  // write an 8 bit tiff with 256 colour colour table, if outTIFFCTab == NULL then do Grayscale
  //
  FILE * fileAccess = NULL;  
  int     imageSize= 0;
  
#ifdef TIFF_BIGENDIAN_OS
  const int tiffPreHeader0= 0x4d4d002a; // basic bigendian TIFF magic number, chose one!
#endif
  
#ifdef TIFF_LITENDIAN_OS
  const int tiffPreHeader0= 0x002a4949; // basic littleendian TIFF magic number
#endif
  
  //
  // fixed header order and offsets.  Preheader, resolution, ctab, tags then data
  //
  const int tiffPreHeader1      = 0x00000618; // offset to tags
  const int tiffXResHeader0     = 0x00000048; // 72 pix/inch default, numerator
  const int tiffXResHeader1     = 0x00000001; // denominator => 72.0 = 72/1
  const int tiffImageDataOffset = 0x000006ae; // specific to this header
  const short nTPS              = 0x000C;     // 12 tags in IFD
  const int ifdTermintator      = 0x00000000; // all TIFF blocks end in 4 bytes of 0
  
  TagStructure tiffTag;
  
  fileAccess = fopen( aFile, "wb");
  if (fileAccess) {
    
    imageSize = (nxOut*nyOut);
    
    zlsio_write( fileAccess, (void*)&tiffPreHeader0, 4UL );
    zlsio_write( fileAccess, (void*)&tiffPreHeader1, 4UL );
    
    zlsio_write( fileAccess, (void*)&tiffXResHeader0, 4UL ); // X-resolution Rational long1/long2
    zlsio_write( fileAccess, (void*)&tiffXResHeader1, 4UL );
    zlsio_write( fileAccess, (void*)&tiffXResHeader0, 4UL ); // Y-resolution Rational long1/long2
    zlsio_write( fileAccess, (void*)&tiffXResHeader1, 4UL );
    
    zlsio_write( fileAccess, outTIFFCTab, sizeof(TIFFCTab) );
    
    zlsio_write( fileAccess, (void*)&nTPS, 2 );
    
    tiffTag.tCode = 0x0100; // width code
    tiffTag.dataType = 4; // long
    tiffTag.nDataValues = 1;
    tiffTag.pDataField = nxOut;
    
    zlsio_write( fileAccess, (void*)&tiffTag, sizeof(TagStructure) );
    
    
    tiffTag.tCode = 0x0101; // height code
    tiffTag.dataType = 4; // long
    tiffTag.nDataValues = 1;
    tiffTag.pDataField = nyOut;
    
    zlsio_write( fileAccess, (void*)&tiffTag, sizeof(TagStructure) );
    
    tiffTag.tCode = 0x0102; // bits per pixel code
    tiffTag.dataType = 3;
    tiffTag.nDataValues = 1;
#ifdef TIFF_BIGENDIAN_OS
    tiffTag.pDataField = 0x00080000; // 8 bits per pixel
#endif
#ifdef TIFF_LITENDIAN_OS
    tiffTag.pDataField = 0x00000008; // 8 bits per pixel
#endif
    
    zlsio_write( fileAccess, (void*)&tiffTag, sizeof(TagStructure) );
    
    tiffTag.tCode = 0x0103; // compression code
    tiffTag.dataType = 3; // short
    tiffTag.nDataValues = 1;
#ifdef TIFF_BIGENDIAN_OS
    tiffTag.pDataField = 0x00010000; //none compression
#endif
#ifdef TIFF_LITENDIAN_OS
    tiffTag.pDataField = 0x00000001; //none compression
#endif
    
    zlsio_write( fileAccess, (void*)&tiffTag, sizeof(TagStructure) );
    
    tiffTag.tCode = 0x0106; // Perception Intent code
    tiffTag.dataType = 3; // short
    tiffTag.nDataValues = 1; 
#ifdef TIFF_BIGENDIAN_OS
    tiffTag.pDataField = 0x00030000; // 1 => 0 = black GS,  3=> colour TABLE
#endif
#ifdef TIFF_LITENDIAN_OS
    tiffTag.pDataField = 0x00000003; // 1 => 0 = black GS,  3=> colour TABLE
#endif
    
    zlsio_write( fileAccess, (void*)&tiffTag, sizeof(TagStructure) );
    
    tiffTag.tCode = 0x0115; // samples per pix
    tiffTag.dataType = 3; // short
    tiffTag.nDataValues = 1;
#ifdef TIFF_BIGENDIAN_OS
    tiffTag.pDataField = 0x00010000; // 1 channel
#endif
#ifdef TIFF_LITENDIAN_OS
    tiffTag.pDataField = 0x00000001; // 1 channel
#endif
    
    zlsio_write( fileAccess, (void*)&tiffTag, sizeof(TagStructure) );
    
    
    tiffTag.tCode = 0x0116; // Rows Per Strip (all in one block!)
    tiffTag.dataType = 4; // long
    tiffTag.nDataValues = 1;
    tiffTag.pDataField = nyOut; // 1 channel
    
    zlsio_write( fileAccess, (void*)&tiffTag, sizeof(TagStructure) );
    
    
    tiffTag.tCode = 0x0117; // StripByteCounts
    tiffTag.dataType = 4; // long
    tiffTag.nDataValues = 1;
    tiffTag.pDataField = imageSize; // 1 channel
    
    zlsio_write( fileAccess, (void*)&tiffTag, sizeof(TagStructure) );
    
    
    tiffTag.tCode = 0x0111; // strip offset from start of whole file
    tiffTag.dataType = 4; // long
    tiffTag.nDataValues = 1;
    tiffTag.pDataField = tiffImageDataOffset;
    
    zlsio_write( fileAccess, (void*)&tiffTag, sizeof(TagStructure) );
    
    tiffTag.tCode = 0x011A; // X-resolution rational offset back to preheader
    tiffTag.dataType = 5;
    tiffTag.nDataValues = 1;
    tiffTag.pDataField = 0x00000008;
    
    zlsio_write( fileAccess, (void*)&tiffTag, sizeof(TagStructure) );
    
    tiffTag.tCode = 0x011B; // Y-resolution rational offset back to preheader
    tiffTag.dataType = 5;
    tiffTag.nDataValues = 1;
    tiffTag.pDataField = 0x00000010;
    
    zlsio_write( fileAccess, (void*)&tiffTag, sizeof(TagStructure) );
    
    tiffTag.tCode = 0x0140; // colour map/table
    tiffTag.dataType = 3;   // SHORT
    tiffTag.nDataValues = 3*256;
    tiffTag.pDataField = 0x00000018; // location of table n bytes from SOF
    zlsio_write( fileAccess, (void*)&tiffTag, sizeof(TagStructure) );
    
    zlsio_write( fileAccess, (void*)&ifdTermintator, 4UL );
    
    // at last- write the data bytes out, and keep main error code
    
    err      = zlsio_write( fileAccess, img, imageSize );
    
    fclose(fileAccess);
    
  } else { err = -1; }
  
  return (err);
  
}
